/*-
 * Copyright (c) 2006 David Xu <davidxu@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <sys/mqueue.h>

#include "namespace.h"
#include <errno.h>
#include <pthread.h>
#include <stddef.h>
#include <signal.h>
#include <unistd.h>
#include "sigev_thread.h"
#include "un-namespace.h"

__weak_reference(___mq_close, mq_close);
__weak_reference(___mq_close, _mq_close);
__weak_reference(__mq_notify, mq_notify);
__weak_reference(__mq_notify, _mq_notify);

extern int __sys_mq_notify(mqd_t mqd, const struct sigevent *evp);
extern int __sys_close(int fd);

int
___mq_close(int mqd)
{
	int ret;

	__sigev_list_lock();
	__sigev_delete(SI_MESGQ, (sigev_id_t)mqd);
	ret = __sys_close(mqd);
	__sigev_list_unlock();
	return (ret);
}

typedef void (*mq_func)(union sigval val, int mqd);

static void
mq_dispatch(struct sigev_node *sn, siginfo_t *si)
{
	mq_func f = sn->sn_func;

	/* I want to avoid expired notification. */
	if (sn->sn_gen == si->si_value.sival_int)
		f(sn->sn_value, si->si_mqd);
}

int
__mq_notify(mqd_t mqd, const struct sigevent *evp)
{
	struct sigevent ev;
	struct sigev_node *sn;
	int ret, err;

	if (__sigev_check_init())
		return (EINVAL);

	__sigev_list_lock();
	__sigev_delete(SI_MESGQ, (sigev_id_t)mqd);
	if (evp == NULL || evp->sigev_notify != SIGEV_THREAD) {
		ret = __sys_mq_notify(mqd, evp);
		__sigev_list_unlock();
		return (ret);
	}
	__sigev_list_unlock();

	sn = __sigev_alloc(SI_MESGQ, evp);
	if (sn == NULL) {
		errno = EAGAIN;
		return (-1);
	}
	sn->sn_id = mqd;
	sn->sn_flags |= SNF_ONESHOT;
	sn->sn_dispatch = mq_dispatch;
	__sigev_get_sigevent(sn, &ev, sn->sn_gen);
	__sigev_list_lock();
	ret = __sys_mq_notify(mqd, &ev);
	if (ret != 0) {
		err = errno;
		__sigev_list_unlock();
		__sigev_free(sn);
		errno = err;
		return (-1);
	}
	__sigev_register(sn);
	__sigev_list_unlock();
	return (0);
}
